# These are exposed functions and objects from the `synapseclient.core.models` package.
# However, these functions and objects are not public APIs for the Synapse Python client.
# The Synapse Engineering team is free to change their signatures and implementations anytime.
# Please use them at your own risk.

from .dict_object import DictObject  # noqa
